#include <torch/extension.h>
#include <ATen/ATen.h>
#include <vector>
#include <string>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

namespace py = pybind11;

// Unified function for batch normalization and ReLU
at::Tensor batch_norm_relu(
    at::Tensor x,
    at::Tensor weight,
    at::Tensor bias,
    at::Tensor running_mean,
    at::Tensor running_var,
    bool training,
    double momentum,
    double eps) {
  x = at::batch_norm(x, weight, bias, running_mean, running_var,
                     training, momentum, eps, true);
  return at::relu(x);
}

// Process a batch of dense layers
template<int BATCH_SIZE = 4>
void process_dense_layers_batch(
    std::vector<at::Tensor>& features,
    at::Tensor& x,
    const py::object& params,
    const std::string& block_prefix,
    int start_idx,
    int remaining_layers,
    bool is_training
) {
    auto get_param = [&](const std::string& key) -> at::Tensor {
        return params.attr("__getitem__")(key.c_str()).cast<at::Tensor>();
    };

    #pragma unroll
    for (int i = 0; i < BATCH_SIZE && i < remaining_layers; ++i) {
        int layer_idx = start_idx + i;
        std::string prefix = block_prefix + std::to_string(layer_idx) + "_";
        
        auto bn_weight = get_param(prefix + "bn_weight");
        auto bn_bias = get_param(prefix + "bn_bias");
        auto bn_mean = get_param(prefix + "bn_mean");
        auto bn_var = get_param(prefix + "bn_var");
        auto conv_weight = get_param(prefix + "conv_weight");

        x = batch_norm_relu(x, bn_weight, bn_bias, bn_mean, bn_var,
                            is_training, 0.1, 1e-5);
        x = at::conv2d(x, conv_weight, /*bias=*/{}, /*stride=*/{1, 1}, /*padding=*/{1, 1});
        x = at::dropout(x, /*p=*/0.0, is_training);
        
        features.push_back(x);
        x = at::cat(features, 1);
    }
}

at::Tensor transition_layer_fn(
    at::Tensor x,
    at::Tensor bn_weight,
    at::Tensor bn_bias,
    at::Tensor bn_mean,
    at::Tensor bn_var,
    at::Tensor conv_weight,
    bool is_training
) {
    x = batch_norm_relu(x, bn_weight, bn_bias, bn_mean, bn_var,
                        is_training, 0.1, 1e-5);
    x = at::conv2d(x, conv_weight);
    x = at::avg_pool2d(x, /*kernel_size=*/{2, 2}, /*stride=*/{2, 2});
    return x;
}

at::Tensor module_fn(
    at::Tensor x,
    py::object params,
    bool is_training
) {
    auto get_param = [&](const std::string& key) -> at::Tensor {
        return params.attr("__getitem__")(key.c_str()).cast<at::Tensor>();
    };

    auto features_conv_weight = get_param("features_conv_weight");
    x = at::conv2d(x, features_conv_weight, /*bias=*/{}, /*stride=*/{2, 2}, /*padding=*/{3, 3});

    auto features_bn_mean = get_param("features_bn_mean");
    auto features_bn_var = get_param("features_bn_var");
    auto features_bn_weight = get_param("features_bn_weight");
    auto features_bn_bias = get_param("features_bn_bias");

    x = batch_norm_relu(x, features_bn_weight, features_bn_bias,
                        features_bn_mean, features_bn_var,
                        is_training, 0.1, 1e-5);
    x = at::max_pool2d(x, /*kernel_size=*/{3, 3}, /*stride=*/{2, 2}, /*padding=*/{1, 1});

    constexpr int BATCH_SIZE = 4;
    std::vector<int> num_layers = {6, 12, 24, 16};

    #pragma unroll
    for (int block = 0; block < 4; ++block) {
        std::vector<at::Tensor> features;
        features.push_back(x);

        for (int layer = 0; layer < num_layers[block]; layer += BATCH_SIZE) {
            std::string block_prefix = "block" + std::to_string(block) + "_layer";
            int remaining_layers = num_layers[block] - layer;
            
            process_dense_layers_batch<BATCH_SIZE>(
                features, x, params, block_prefix,
                layer, remaining_layers, is_training
            );
        }

        if (block != 3) {
            std::string prefix = "transition" + std::to_string(block) + "_";
            auto bn_weight = get_param(prefix + "bn_weight");
            auto bn_bias = get_param(prefix + "bn_bias");
            auto bn_mean = get_param(prefix + "bn_mean");
            auto bn_var = get_param(prefix + "bn_var");
            auto conv_weight = get_param(prefix + "conv_weight");

            x = transition_layer_fn(x, bn_weight, bn_bias, bn_mean,
                                  bn_var, conv_weight, is_training);
        }
    }

    auto final_bn_mean = get_param("final_bn_mean");
    auto final_bn_var = get_param("final_bn_var");
    auto final_bn_weight = get_param("final_bn_weight");
    auto final_bn_bias = get_param("final_bn_bias");

    x = batch_norm_relu(x, final_bn_weight, final_bn_bias,
                        final_bn_mean, final_bn_var,
                        is_training, 0.1, 1e-5);
    x = at::adaptive_avg_pool2d(x, {1, 1}).reshape({x.size(0), -1});

    auto classifier_weight = get_param("classifier_weight");
    auto classifier_bias = get_param("classifier_bias");
    x = at::linear(x, classifier_weight, classifier_bias);

    return x;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn, "Optimized DenseNet121 forward");
}
